/*********************************************************************
 This code is based on an example for the Bluefruit LE modules
 by Adafruit

 MIT license, check LICENSE for more information
 All text above, and the splash screen below must be included in
 any redistribution
*********************************************************************/

#include <Arduino.h>
#include <SPI.h>

#include "Adafruit_Si7021.h"

#include "Adafruit_BLE.h"
#include "Adafruit_BluefruitLE_SPI.h"

// Use the LowPower Library to reduce energy of the 32u4 during sleep state
#include "LowPower.h"

#include "BluefruitConfig.h"

// Create the bluefruit object, either software serial...uncomment these lines
/* ...hardware SPI, using SCK/MOSI/MISO hardware SPI pins and then user selected CS/IRQ/RST */
Adafruit_BluefruitLE_SPI ble(BLUEFRUIT_SPI_CS, BLUEFRUIT_SPI_IRQ, BLUEFRUIT_SPI_RST);


// Use Adafruits Si7021 temperature and humidity sensor
Adafruit_Si7021 sensor = Adafruit_Si7021();

// Define the pin that is used on feather board to measure battery voltage
#define VBATPIN A9

// A small helper
void error(int speed) {
  while ( true ) {
    digitalWrite( 13, ! digitalRead( 13 ) ) ;
    delay( speed );
  }
}

void notify(int speed) {
  for(int i=0; i<30;i++) {
    digitalWrite( 13, ! digitalRead( 13 ) ) ;
    delay( speed );
  }
}

/* The service information */

// Define the Environmental Sensing Service
int32_t esServiceId;
int32_t esTemperatureCharId;
int32_t esHumidityCharId;
int32_t esMagneticDeclination;

// Define update counter for GAP advertising and the number of cycles before
// GAP advertising is updated
int updateCounter;
#define NUMCYCLES 8

/**************************************************************************/
/*!
    @brief  Sets up the HW an the BLE module (this function is called
            automatically on startup)
*/
/**************************************************************************/
void setup(void)
{
  delay(500);

  boolean success;

  if ( !ble.begin(false) )
  {
    error(100);
  }

  /* Perform a factory reset to make sure everything is in a known state */
  if (! ble.factoryReset() ){
       error(100);
  }

  /* Disable command echo from Bluefruit */
  ble.echo(false);
  ble.verbose(false);

  /* Change the device name to make it easier to find */
  if (! ble.sendCommandCheckOK(F("AT+GAPDEVNAME=Bluefruit TH1")) ) {
    error(100);
  }

  /* Add the environmental sensing service definition */
  /* Service ID should be 1 */
  success = ble.sendCommandWithIntReply( F("AT+GATTADDSERVICE=UUID=0x181A"), &esServiceId);
  if (! success) {
    error(100);
  }

  /* Add the Temperature Measurement characteristic */
  /* Chars ID for Measurement should be 1 */
  success = ble.sendCommandWithIntReply( F("AT+GATTADDCHAR=UUID=0x2A6E, PROPERTIES=0x02, MIN_LEN=2, VALUE=2000, DATATYPE=3"), &esTemperatureCharId);
    if (! success) {
    error(100);
  }

  // Add the Humidity Measurement characteristic
  // Chars ID for Body should be 2
  success = ble.sendCommandWithIntReply( F("AT+GATTADDCHAR=UUID=0x2A6F, PROPERTIES=0x02, MIN_LEN=2, VALUE=5000"), &esHumidityCharId);
    if (! success) {
    error(100);
  }

  // Misuse the  the magnetic declination characteristic to measure battery voltage
  // Chars ID for Body should be 3
  success = ble.sendCommandWithIntReply( F("AT+GATTADDCHAR=UUID=0x2A2C, PROPERTIES=0x02, MIN_LEN=2, VALUE=3700"), &esMagneticDeclination);
    if (! success) {
    error(100);
  }

  /* Add the environmental sensing Service to the advertising data (needed for Nordic apps to detect the service) */
  // 02-01-06 -> 2 Bytes , Datatype = 01 => Flags, Value = 06 = 0b00000110 => LE General Discoverable Mode / BR/EDR Not Supported
  // 05-02-1a-18-0a-18 -> 5 Bytes, Datatype = 02 => Incomplete List of 16-bit Service Class UUID, Values = 0x181A, 0x180A
  // 09-FF-00-00-01-02-03-04-05-06   -> 9 Bytes, Datatype = FF => Manufacturer Specific Data, Value 00, 00, 01, 02, 03, 04, 05, 06
  // But take care! the first two bytes of manufacturing data are manufacturer id: https://www.bluetooth.com/specifications/assigned-numbers/company-Identifiers
  ble.sendCommandCheckOK( F("AT+GAPSETADVDATA=02-01-06-05-02-1a-18-0a-18-09-FF-00-00-01-02-03-04-05-06") );

  /*
   * The manufacturer specific data field can be updated to broadcast actual data.
   * However, the device needs always a reset afterwards!
   */

  /* Reset the device for the new service setting changes to take effect */
  ble.reset();
  /* Disable command echo from Bluefruit */
  ble.echo(false);
  ble.verbose(false);

  // Disable MOD LED in order to save battery power
  // Normal mode is: MODE
  if (!ble.sendCommandCheckOK( F("AT+HWMODELED=DISABLE") ) ) {
    error(100);
  }
  
  // Start the sensor reading on the I2C bus
  sensor.begin();

  updateCounter = NUMCYCLES;
}

/** Send randomized heart rate data continuously **/
void loop(void)
{
  /* Delay before next measurement update */
  // Enter idle state for 8 s with the rest of peripherals turned off
  LowPower.powerDown(SLEEP_8S, ADC_OFF, BOD_OFF);

  while( ble.isConnected() );
      
  short  temperature = (short)(sensor.readTemperature() * 100.0 + 0.5);
  unsigned short  humidity =  (unsigned short)(sensor.readHumidity() * 100.0 +0.5);
  float measuredvbat = analogRead(VBATPIN);
  measuredvbat *= 2;    // we divided by 2, so multiply back
  measuredvbat *= 3.3;  // Multiply by 3.3V, our reference voltage
  measuredvbat /= 1024; // convert to voltage

  short  batVoltage = (short)(measuredvbat * 100.0);

  /* Command is sent when \n (\r) or println is called */
  /* AT+GATTCHAR=CharacteristicID,value */
  ble.print( F("AT+GATTCHAR=") );
  ble.print( esTemperatureCharId );
  ble.print( F(",") );
  ble.println(temperature);

  ble.print( F("AT+GATTCHAR=") );
  ble.print( esHumidityCharId );
  ble.print( F(",") );
  ble.println(humidity);

  ble.print( F("AT+GATTCHAR=") );
  ble.print( esMagneticDeclination );
  ble.print( F(",") );
  ble.println(batVoltage);

  /* Check if command executed OK */
  if ( !ble.waitForOK() )
  {
    notify(100);
    ble.reset();
    ble.echo(false);
    ble.verbose(false);
  }

  if(updateCounter > NUMCYCLES) {
    //notify(20);
    updateGAPAdvertising(temperature, humidity, batVoltage);
    updateCounter = 1;
    //notify(30);
  }
  else {
    updateCounter++;
  }
  
}

void updateGAPAdvertising(short temperature, unsigned short humidity, short batVoltage) {
  char tmp[5],tmp2[6], advData[80];
  strcpy(advData, "AT+GAPSETADVDATA=02-01-06-05-02-1a-18-0a-18-09-FF-00-00-");

  sprintf(tmp, "%.4X",temperature);
  prepareHex(tmp, tmp2);
  strcat(advData, tmp2);
  strcat(advData, "-");

  sprintf(tmp, "%.4X",humidity);
  prepareHex(tmp, tmp2);
  strcat(advData, tmp2);
  strcat(advData, "-");

  sprintf(tmp, "%.4X",batVoltage);
  prepareHex(tmp, tmp2);
  strcat(advData, tmp2);
 
  ble.sendCommandCheckOK( advData );
  /* Reset the device for the new service setting changes to take effect */
  while( ble.isConnected() );
  ble.reset();
  /* Disable command echo from Bluefruit */
  ble.echo(false);
  ble.verbose(false);
  
  // Disable MOD LED in order to save battery power
  // Normal mode is: MODE
  ble.sendCommandCheckOK( F("AT+HWMODELED=DISABLE") );

  ble.waitForOK();
  
}

void prepareHex(char * input, char *result) {
  strncpy(result,input,2);
  result[2] = '\0';
  strcat(result,"-");
  strcat(result,input+2);
}
